import logging

logger = logging.getLogger(__name__)


class Report:
    def __init__(self, package):
        self._database = package.imports
        self.exit_status = 0

    def print_report(self):
        logger.debug("Package requirements: %s", self._database._requirements)
        logger.debug("Package extras: %s", self._database._extras_requirements)
        logger.debug(
            "User defined mappings: %s",
            self._database.user_mappings,
        )
        self.missing_requirements()
        self.missing_test_requirements()
        self.unneeded_requirements()
        self.requirements_that_should_be_test_requirements()
        self.unneeded_test_requirements()
        self.print_notice()

    def missing_requirements(self):
        self._print_metric(
            "Missing requirements",
            self._database.get_missing_imports,
        )

    def missing_test_requirements(self):
        self._print_metric(
            "Missing test requirements",
            self._database.get_missing_test_imports,
        )

    def unneeded_requirements(self):
        self._print_metric(
            "Unneeded requirements",
            self._database.get_unneeded_requirements,
        )

    def requirements_that_should_be_test_requirements(self):
        self._print_metric(
            "Requirements that should be test requirements",
            self._database.requirements_that_should_be_test_requirements,
        )

    def unneeded_test_requirements(self):
        self._print_metric(
            "Unneeded test requirements",
            self._database.get_unneeded_test_requirements,
        )

    @staticmethod
    def print_notice():
        print("")
        print("Note: requirements are taken from the egginfo dir, so you need")
        print("to re-generate it, via `python -m build`, setuptools, zc.buildout")
        print("or any other means.")
        print("")

    def _print_metric(self, title, method):
        missed = method()
        if len(missed) == 0:
            return

        self.exit_status = 1

        self._print_header(title)
        for dotted_name in missed:
            print(f"     {dotted_name.name}")

    @staticmethod
    def _print_header(message):
        if message:
            print("")
            print(message)
            print("=" * len(message))



import os

from z3c.dependencychecker.package import Package, PackageMetadata
from z3c.dependencychecker.tests.utils import (
    get_extras_requirements_names,
    get_requirements_names,
    get_requirements_names_for_extra,
    get_sorted_imports_paths,
    write_source_file_at,
)


def test_package_has_metadata(minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)

    assert bool(getattr(package, "metadata", False))
    assert isinstance(package.metadata, PackageMetadata)


def test_declared_dependencies(minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    package.set_declared_dependencies()

    requirements = get_requirements_names(package.imports)
    assert len(requirements) == 2
    assert "one" in requirements
    assert "two" in requirements


def test_declared_dependencies_empty(minimal_structure):
    path, package_name = minimal_structure
    requires_file_path = os.path.join(
        path,
        f"{package_name}.egg-info",
        "requires.txt",
    )
    with open(requires_file_path, "w") as requires:
        requires.write("")

    package = Package(path)
    package.set_declared_dependencies()

    assert len(get_requirements_names(package.imports)) == 0


def test_declared_extras_dependencies_one_extra(minimal_structure):
    path, package_name = minimal_structure
    requires_file_path = os.path.join(
        path,
        f"{package_name}.egg-info",
        "requires.txt",
    )
    with open(requires_file_path, "w") as requires:
        requires.write("\n".join(["[extra]", "my.package", "another.package"]))

    package = Package(path)
    package.set_declared_dependencies()
    package.set_declared_extras_dependencies()

    extras_names = get_extras_requirements_names(package.imports)
    assert len(extras_names) == 1
    assert "extra" in extras_names
    names = get_requirements_names_for_extra(package.imports, extra="extra")
    assert "another.package" in names
    assert "my.package" in names


def test_declared_extras_dependencies_only_on_extras(minimal_structure):
    """Check that main dependencies are not bundled on the extras"""
    path, package_name = minimal_structure
    requires_file_path = os.path.join(
        path,
        f"{package_name}.egg-info",
        "requires.txt",
    )
    with open(requires_file_path, "w") as requires:
        requires.write(
            "\n".join(["setuptools", "", "[extra]", "my_package", "my_other_package"])
        )

    package = Package(path)
    package.set_declared_dependencies()
    package.set_declared_extras_dependencies()

    names = get_requirements_names_for_extra(package.imports, extra="extra")
    assert "setuptools" not in names


def test_multiple_extras(minimal_structure):
    path, package_name = minimal_structure
    requires_file_path = os.path.join(
        path,
        f"{package_name}.egg-info",
        "requires.txt",
    )
    with open(requires_file_path, "w") as requires:
        requires.write(
            "\n".join(
                [
                    "setuptools",
                    "",
                    "[extra]",
                    "my_package",
                    "my_other_package",
                    "[second]",
                    "just",
                    "laughs",
                ]
            )
        )

    package = Package(path)
    package.set_declared_extras_dependencies()

    extras_names = get_extras_requirements_names(package.imports)
    assert len(extras_names) == 2
    assert "second" in extras_names
    assert "extra" in extras_names


def test_no_python_modules_found(minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    package.analyze_package()
    assert len(package.imports.imports_used) == 0


def test_python_modules_on_top_level_sources(minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at((path, package_name), "__init__.py")
    package = Package(path)
    package.analyze_package()

    assert len(package.imports.imports_used) == 1
    paths = get_sorted_imports_paths(package.imports)
    assert paths[0].endswith("__init__.py")


def test_python_modules_found_inner_folders(minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at((path, package_name), "__init__.py")
    write_source_file_at((path, package_name, "test"), "__init__.py")
    write_source_file_at((path, package_name, "test"), "test.py")

    package = Package(path)
    package.analyze_package()
    paths = get_sorted_imports_paths(package.imports)
    assert len(paths) == 3
    assert paths[0].endswith("__init__.py")
    assert paths[1].endswith("__init__.py")
    assert paths[2].endswith("test.py")


def test_top_level_is_a_file(minimal_structure):
    path, package_name = minimal_structure

    egg_info_folder = os.path.join(path, f"{package_name}.egg-info")
    top_level_file_path = os.path.join(egg_info_folder, "top_level.txt")
    with open(top_level_file_path) as top_level_file:
        top_level_folder = top_level_file.read().strip()

    top_level_sources = os.path.join(path, top_level_folder)
    os.removedirs(top_level_sources)
    with open(f"{top_level_sources}.py", "w") as top_level_file:
        top_level_file.write("import one")

    package = Package(path)
    package.analyze_package()
    paths = get_sorted_imports_paths(package.imports)
    assert len(paths) == 1
    assert paths[0].endswith(f"{top_level_folder}.py")



import os

import pytest

from z3c.dependencychecker.modules import ZCMLFile
from z3c.dependencychecker.tests.utils import write_source_file_at

ZCML_TEMPLATE = """
<configure
  xmlns="http://namespaces.zope.org/zope"
  xmlns:browser="http://namespaces.zope.org/browser"
  xmlns:genericsetup="http://namespaces.zope.org/genericsetup">

{0}

</configure>
"""

RELATIVE_IMPORT = ".IRuleAssignable"
ABSOLUTE_IMPORT_1 = "plone.interfaces.IContent"
ABSOLUTE_IMPORT_2 = "zope.interfaces.IFolder"
ASTERISK_IMPORT = "*"

# list all possible combinations of how the imports can look like
# and how many of them are expected to be considered valid imports
# by z3c.dependencychecker
TEST_IMPORTS_MATRIX = (
    (0, f"{RELATIVE_IMPORT}"),
    (0, f"{ASTERISK_IMPORT}"),
    (0, f"{ASTERISK_IMPORT} {RELATIVE_IMPORT}"),
    (1, f"{ABSOLUTE_IMPORT_1}"),
    (1, f"{ABSOLUTE_IMPORT_2} {RELATIVE_IMPORT}"),
    (1, f"{ASTERISK_IMPORT} {ABSOLUTE_IMPORT_1}"),
    (2, f"{ABSOLUTE_IMPORT_1} {ABSOLUTE_IMPORT_2}"),
    (2, f"{ABSOLUTE_IMPORT_1} {ABSOLUTE_IMPORT_2} {ASTERISK_IMPORT}"),
)


def _get_zcml_imports_on_file(folder, source):
    full_content = ZCML_TEMPLATE.format(source)
    temporal_file = write_source_file_at(
        (folder.strpath,),
        source_code=full_content,
        filename="configure.zcml",
    )

    zcml_file = ZCMLFile(folder.strpath, temporal_file)
    dotted_names = [x.name for x in zcml_file.scan()]
    return dotted_names


def _verify_dotted_names(dotted_names, imports, found):
    assert len(dotted_names) == found
    for an_import in imports.split(" "):
        if an_import in (RELATIVE_IMPORT, ASTERISK_IMPORT):
            continue
        assert an_import in dotted_names


def test_create_from_files_nothing(minimal_structure):
    path, package_name = minimal_structure
    modules_found = list(ZCMLFile.create_from_files(path))
    assert len(modules_found) == 0


def test_create_from_files_deep_nested(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, "src")
    write_source_file_at([src_path, "a", "b"], filename="configure.zcml")

    modules_found = list(ZCMLFile.create_from_files(src_path))
    assert len(modules_found) == 1


@pytest.mark.parametrize("result,imports", TEST_IMPORTS_MATRIX)
def test_zcml_include(tmpdir, result, imports):
    """Check that the ZCML <include directive is scanned."""
    zcml_stanza = f'<include package="{imports}" />'
    dotted_names = _get_zcml_imports_on_file(tmpdir, zcml_stanza)
    _verify_dotted_names(dotted_names, imports, result)


@pytest.mark.parametrize("attribute", ("for", "factory", "provides"))
@pytest.mark.parametrize("result,imports", TEST_IMPORTS_MATRIX)
def test_zcml_adapter(tmpdir, result, imports, attribute):
    """Check that the ZCML <adapter directive is scanned."""
    zcml_stanza = f'<adapter {attribute}="{imports}" />'
    dotted_names = _get_zcml_imports_on_file(tmpdir, zcml_stanza)
    _verify_dotted_names(dotted_names, imports, result)


@pytest.mark.parametrize("attribute", ("provides", "component"))
@pytest.mark.parametrize("result,imports", TEST_IMPORTS_MATRIX)
def test_zcml_utility(tmpdir, result, imports, attribute):
    """Check that the ZCML <utility directive is scanned."""
    zcml_stanza = f'<utility {attribute}="{imports}" />'
    dotted_names = _get_zcml_imports_on_file(tmpdir, zcml_stanza)
    _verify_dotted_names(dotted_names, imports, result)


@pytest.mark.parametrize("attribute", ("class", "for", "layer"))
@pytest.mark.parametrize("result,imports", TEST_IMPORTS_MATRIX)
def test_zcml_browser_page(tmpdir, result, imports, attribute):
    """Check that the ZCML <browser:page directive is scanned."""
    zcml_stanza = f'<browser:page {attribute}="{imports}" />'
    dotted_names = _get_zcml_imports_on_file(tmpdir, zcml_stanza)
    _verify_dotted_names(dotted_names, imports, result)


@pytest.mark.parametrize("attribute", ("handler", "for"))
@pytest.mark.parametrize("result,imports", TEST_IMPORTS_MATRIX)
def test_zcml_subscriber(tmpdir, result, imports, attribute):
    """Check that the ZCML <subscriber directive is scanned."""
    zcml_stanza = f'<subscriber {attribute}="{imports}" />'
    dotted_names = _get_zcml_imports_on_file(tmpdir, zcml_stanza)
    _verify_dotted_names(dotted_names, imports, result)


@pytest.mark.parametrize("result,imports", TEST_IMPORTS_MATRIX)
def test_zcml_security_policy(tmpdir, result, imports):
    """Check that the ZCML <securityPolicy directive is scanned."""
    zcml_stanza = f'<securityPolicy component="{imports}" />'
    dotted_names = _get_zcml_imports_on_file(tmpdir, zcml_stanza)
    _verify_dotted_names(dotted_names, imports, result)


@pytest.mark.parametrize("result,imports", TEST_IMPORTS_MATRIX)
def test_zcml_genericsetup(tmpdir, result, imports):
    """Check that the ZCML <genericsetup:registerProfile directive is scanned."""
    zcml_stanza = f'<genericsetup:registerProfile provides="{imports}" />'
    dotted_names = _get_zcml_imports_on_file(tmpdir, zcml_stanza)
    _verify_dotted_names(dotted_names, imports, result)


@pytest.mark.parametrize("result,imports", TEST_IMPORTS_MATRIX)
def test_zcml_implements(tmpdir, result, imports):
    """Check that the ZCML <implements directive is scanned."""
    zcml_stanza = f'<implements interface="{imports}" />'
    dotted_names = _get_zcml_imports_on_file(tmpdir, zcml_stanza)
    _verify_dotted_names(dotted_names, imports, result)



import pytest

from z3c.dependencychecker.dotted_name import DottedName


def test_fallback():
    test_import = object()
    requirement = DottedName("plone.app.dexterity")
    assert test_import not in requirement


@pytest.mark.parametrize(
    "test_import,requirement,is_contained",
    (
        # umbrella distributions should never match
        ("Plone", "Plone", False),
        ("Zope", "Zope", False),
        # all other names match with themselves
        ("z3c", "z3c", True),
        ("Plone", "Zope", False),
        ("plo", "plone.app.imaging", False),
        ("reinout.happy", "re", False),
        ("ima", "plone.app.imaging", False),
        ("imaging", "plone.app.imaging", False),
        ("plone.app.imaging.interfaces.IImage", "plone.app.imaging", True),
        ("plone.app.dexterity", "plone.app.imaging", False),
        ("plone.app.dexterity.interfaces", "plone.app.imaging", False),
        ("plone.app.dexterity", "plone.app.imaging.interfaces", False),
    ),
)
def test_containment(test_import, requirement, is_contained):
    """Check DottedName.__contains__ dunder logic"""
    test_dotted_name = DottedName(test_import)
    requirement_dotted_name = DottedName(requirement)
    if is_contained:
        assert test_dotted_name in requirement_dotted_name
    else:
        assert test_dotted_name not in requirement_dotted_name



from unittest import mock

import pytest

from z3c.dependencychecker.dotted_name import DottedName


def test_minimal():
    obj = DottedName("dotted.name")
    assert obj.name == "dotted.name"


def test_no_path():
    obj = DottedName("dotted.name")
    assert obj.file_path is None


def test_file_path_given():
    obj = DottedName("dotted.name", file_path="/one/two")
    assert obj.file_path == "/one/two"


def test_requirement_constructor():
    mock_object = mock.MagicMock()
    mock_property = mock.PropertyMock(return_value="my.dotted.name")
    type(mock_object).project_name = mock_property

    obj = DottedName.from_requirement(mock_object)

    assert obj.file_path is None
    assert obj.name == "my.dotted.name"


def test_requirement_constructor_with_path():
    mock_object = mock.MagicMock()
    mock_property = mock.PropertyMock(return_value="my.dotted.name")
    type(mock_object).project_name = mock_property

    obj = DottedName.from_requirement(mock_object, file_path="/one/two")

    assert obj.file_path == "/one/two"


def test_is_test_import():
    obj = DottedName("plone.app.dexterity", is_test=True)
    assert obj.is_test


def test_not_a_test_import():
    obj = DottedName("plone.app.dexterity")
    assert not obj.is_test


def test_safe_name_no_changes():
    obj = DottedName("hi.there")
    assert obj.name == obj.safe_name


def test_safe_name_lowercase():
    original_name = "Hi.There"
    obj = DottedName(original_name)
    assert original_name.lower() == obj.safe_name


def test_safe_name_dash_to_underscore():
    original_name = "hi-there"
    obj = DottedName(original_name)
    assert original_name.replace("-", "_") == obj.safe_name


def test_safe_name_all_at_once():
    original_name = "Hi-There"
    obj = DottedName(original_name)
    assert obj.safe_name == "hi_there"


def test_namespaces_none():
    obj = DottedName("plain_package_name")
    assert obj.namespaces == ["plain_package_name"]


def test_namespaces_some():
    obj = DottedName("plone.app.dexterity")
    assert len(obj.namespaces) == 3


def test_namespaces_some_details():
    obj = DottedName("plone.app.dexterity")
    assert obj.namespaces == ["plone", "app", "dexterity"]


def test_namespaces_really_long():
    obj = DottedName("one.two.three.four.five.six")
    assert len(obj.namespaces) == 6
    assert obj.namespaces == ["one", "two", "three", "four", "five", "six"]


def test_is_namespaced_not():
    obj = DottedName("nope")
    assert obj.is_namespaced is False


def test_is_namespaced():
    obj = DottedName("plone.app.dexterity")
    assert obj.is_namespaced


def test_comparison_fallback():
    obj1 = DottedName("plone.app.dexterity")
    with pytest.raises(TypeError):
        obj1 < 33  # noqa: B015


def test_comparison_bigger_than():
    obj1 = DottedName("plone.app.ldap")
    obj2 = DottedName("plone.app.dexterity")
    assert obj1 > obj2


def test_comparison_bigger_or_equal_than():
    obj1 = DottedName("plone.app.ldap")
    obj2 = DottedName("plone.app.dexterity")
    assert obj1 >= obj2
    assert obj1 >= obj1


def test_comparison_smaller_than():
    obj1 = DottedName("plone.app.dexterity")
    obj2 = DottedName("plone.app.ldap")
    assert obj1 < obj2


def test_comparison_smaller_or_equal_than():
    obj1 = DottedName("plone.app.dexterity")
    obj2 = DottedName("plone.app.ldap")
    assert obj1 <= obj2
    assert obj1 <= obj1


def test_comparison_equality():
    obj1 = DottedName("plone.app.dexterity")
    assert obj1 == obj1


def test_comparison_equality_fallback():
    obj1 = DottedName("plone.app.dexterity")
    result = obj1 == object()
    assert not result


def test_repr():
    obj = DottedName("z3c.dependencychecker")
    assert repr(obj) == "<DottedName z3c.dependencychecker>"


def test_hash():
    obj1 = DottedName("bla")
    assert isinstance(hash(obj1), int)


def test_hash_in_use():
    obj1 = DottedName("one")
    obj2 = DottedName("two")
    obj3 = DottedName("three")
    obj4 = DottedName("one")
    uniques = {obj1, obj2, obj3, obj4}
    assert len(uniques) == 3



import ast
import fnmatch
import logging
import os
import re
from xml.etree import ElementTree

from z3c.dependencychecker.dotted_name import DottedName

TEST_REGEX = r"""
{0}    # Needs to start with the path separator
\w*    # There could be some characters (or none)
test   # has to contain test on it
s?     # ensure plurals are handled as well
(      # Needs to end with either a path or a file extension:
{0}    # path separator
|      # OR
\w*    # some possible characters
\.     # a dot
\w+    # extension
)
""".format(
    os.sep
)

TEST_IN_PATH_REGEX = re.compile(TEST_REGEX, re.VERBOSE)

logger = logging.getLogger(__name__)

FOLDERS_TO_IGNORE = ("node_modules", "__pycache__", "venv")


class BaseModule:
    def __init__(self, package_path, full_path):
        self.path = full_path
        self._relative_path = self._get_relative_path(package_path, full_path)
        self.testing = self._is_test_module()

    @staticmethod
    def _get_relative_path(package_path, full_path):
        return full_path[len(package_path) :]

    def _is_test_module(self):
        return bool(re.search(TEST_IN_PATH_REGEX, self._relative_path))

    @classmethod
    def create_from_files(cls, top_dir):
        raise NotImplementedError

    def scan(self):
        raise NotImplementedError

    def walk_and_filter_folder(folder):
        for path, folders, filenames in os.walk(folder):
            # ignore known folder names
            folders[:] = [d for d in folders if d not in FOLDERS_TO_IGNORE]
            # ignore all folders that start with a dot
            folders[:] = fnmatch.filter(folders, "[!.]*")
            for filename in filenames:
                yield path, filename


class PythonModule(BaseModule):
    @classmethod
    def create_from_files(cls, top_dir):
        """Find all python files in the package

        For that it gets the path to where top_level.txt points to,
        which is not always a folder:
        - folder example: z3c.dependencychecker itself
        - file example: flake8-isort or any other single file distribution

        Return this very same class, which would allow to call the scan()
        method to get an iterator over all this file's imports.
        """
        if top_dir.endswith(".py"):
            yield cls(top_dir, top_dir)
            return

        for path, filename in cls.walk_and_filter_folder(top_dir):
            if filename.endswith(".py"):
                yield cls(
                    top_dir,
                    os.path.join(path, filename),
                )

    def scan(self):
        for node in ast.walk(self._get_tree()):
            yield from self._process_ast_node(node)

    def _get_tree(self):
        with open(self.path) as module_file:
            source_text = module_file.read()
        tree = ast.parse(source_text)
        return tree

    def _process_ast_node(self, node):
        if isinstance(node, ast.Import):
            for name in node.names:
                dotted_name = name.name
                yield DottedName(
                    dotted_name,
                    file_path=self.path,
                    is_test=self.testing,
                )

        elif isinstance(node, ast.ImportFrom):
            if self._is_relative_import(node):
                return
            for name in node.names:
                if name.name == "*":
                    dotted_name = node.module
                else:
                    dotted_name = f"{node.module}.{name.name}"
                yield DottedName(
                    dotted_name,
                    file_path=self.path,
                    is_test=self.testing,
                )

    @staticmethod
    def _is_relative_import(import_node):
        return import_node.level > 0


class ZCMLFile(BaseModule):
    """Extract imports from .zcml files

    These files are in common use in Zope/Plone based projects to define its
    components and more.
    """

    ELEMENTS = {
        "include": ("package",),
        "adapter": ("for", "factory", "provides"),
        "utility": ("provides", "component"),
        "browser:page": ("class", "for", "layer"),
        "subscriber": ("handler", "for"),
        "securityPolicy": ("component",),
        "genericsetup:registerProfile": ("provides",),
        "implements": ("interface",),
    }

    @classmethod
    def create_from_files(cls, top_dir):
        """Find all ZCML files in the package

        Return this very same class, which would allow to call the scan()
        method to get an iterator over all this file's imports.
        """
        if top_dir.endswith(".py"):
            return

        for path, filename in cls.walk_and_filter_folder(top_dir):
            if filename.endswith(".zcml"):
                yield cls(
                    top_dir,
                    os.path.join(path, filename),
                )

    def scan(self):
        tree = ElementTree.parse(self.path).getroot()

        for element in self.ELEMENTS:
            element_namespaced = self._build_namespaced_element(element)
            for node in tree.iter(element_namespaced):
                for attrib in self.ELEMENTS[element]:
                    yield from self._extract_dotted_name(node, attrib)

    def _extract_dotted_name(self, node, attr):
        if attr in node.keys():
            candidate_text = node.get(attr)
            for dotted_name in candidate_text.split(" "):
                if not dotted_name or dotted_name.startswith("."):
                    continue

                if dotted_name == "*":
                    continue

                yield DottedName(
                    dotted_name,
                    file_path=self.path,
                    is_test=self.testing,
                )

    @staticmethod
    def _build_namespaced_element(element):
        if ":" in element:
            namespace, name = element.split(":")
            return "{{http://namespaces.zope.org/{0}}}{1}".format(
                namespace,
                name,
            )

        return f"{{http://namespaces.zope.org/zope}}{element}"


class FTIFile(BaseModule):
    """Extract imports from Factory Type Information files

    These are xml files, usually located inside a types folder and used by
    Zope/Plone based projects to define its content types.
    """

    TYPES_FOLDER = f"{os.sep}types"

    @classmethod
    def create_from_files(cls, top_dir):
        """Find all FTI files, which are xml, in the package

        Return this very same class, which would allow to call the scan()
        method to get an iterator over all this file's imports.
        """
        if top_dir.endswith(".py"):
            return

        for path, filename in cls.walk_and_filter_folder(top_dir):
            if filename.endswith(".xml") and cls.TYPES_FOLDER in path:
                yield cls(
                    top_dir,
                    os.path.join(path, filename),
                )

    def scan(self):
        tree = ElementTree.parse(self.path).getroot()

        for node in tree.iter("property"):
            if "name" in node.keys():
                name = node.get("name")
                if name in ("klass", "schema"):
                    if node.text:
                        yield DottedName(
                            node.text.strip(),
                            file_path=self.path,
                            is_test=self.testing,
                        )


class GSMetadata(BaseModule):
    """Extract imports from Generic Setup metadata.xml files

    These files are in common use in Zope/Plone to define Generic Setup
    profile dependencies between projects.
    """

    PROFILE_RE = re.compile(r"profile-(?P<dotted_name>[\w.]+):[\w\W]+")

    @classmethod
    def create_from_files(cls, top_dir):
        """Find all metadata.xml files in the package

        Return this very same class, which would allow to call the scan()
        method to get an iterator over all this file's imports.
        """
        if top_dir.endswith(".py"):
            return

        for path, filename in cls.walk_and_filter_folder(top_dir):
            if filename == "metadata.xml":
                yield cls(
                    top_dir,
                    os.path.join(path, filename),
                )

    def scan(self):
        tree = ElementTree.parse(self.path).getroot()

        for node in tree.iter("dependency"):
            if not node.text:
                continue
            result = self.PROFILE_RE.search(node.text.strip())
            if result:
                yield DottedName(
                    result.group("dotted_name"),
                    file_path=self.path,
                    is_test=self.testing,
                )


class PythonDocstrings(PythonModule):
    """Extract imports from docstrings found inside python modules

    There are some projects that rather than having separate files to keep the
    tests, they add the inline with the code as docstrings,
    either at class or method/function level.
    """

    NODES_WITH_DOCSTRINGS = (
        ast.Module,
        ast.ClassDef,
        ast.FunctionDef,
    )

    def scan(self):
        for node in ast.walk(self._get_tree()):
            if isinstance(node, self.NODES_WITH_DOCSTRINGS):
                docstring = ast.get_docstring(node)
                yield from self._parse_docstring(docstring)

    def _parse_docstring(self, docstring):
        if not docstring:
            return

        for line in docstring.split("\n"):
            code = self._extract_code(line)
            if code:
                try:
                    tree = ast.parse(code)
                except SyntaxError:
                    logger.debug(
                        'Could not parse "%s" in %s',
                        line,
                        self.path,
                    )
                    continue

                for node in ast.walk(tree):
                    for dotted_name in self._process_ast_node(node):
                        dotted_name.is_test = True
                        yield dotted_name

    @staticmethod
    def _extract_code(line):
        if ">>>" in line:
            position = line.find(">>>") + 3
            return line[position:].strip()


class DocFiles(PythonDocstrings):
    """Extract imports from documentation-like documents

    One extended testing pattern (at least in Zope/Plone based projects) is to
    write documentation in txt/rst files that contain tests to assert that
    documentation.
    """

    @classmethod
    def create_from_files(cls, top_dir):
        """Find all documentation files in the package

        For that it gets the path to where top_level.txt points to,
        which is not always a folder:
        - folder example: z3c.dependencychecker itself
        - file example: flake8-isort or any other single file distribution

        Return this very same class, which would allow to call the scan()
        method to get an iterator over all this file's imports.
        """
        if top_dir.endswith(".py"):
            return

        for path, folders, filenames in os.walk(top_dir):
            folders[:] = [d for d in folders if d not in FOLDERS_TO_IGNORE]
            for filename in filenames:
                if filename.endswith(".txt") or filename.endswith(".rst"):
                    yield cls(
                        top_dir,
                        os.path.join(path, filename),
                    )

    def scan(self):
        with open(self.path) as doc_file:
            for line in doc_file:
                code = self._extract_code(line)
                if code:
                    try:
                        tree = ast.parse(code)
                    except SyntaxError:
                        logger.debug(
                            'Could not parse "%s" in %s',
                            line,
                            self.path,
                        )
                        continue

                    for node in ast.walk(tree):
                        for dotted_name in self._process_ast_node(node):
                            dotted_name.is_test = True
                            yield dotted_name


class DjangoSettings(PythonModule):
    """Extract imports from Django settings.py

    These files are used to enable Django components.
    """

    @classmethod
    def create_from_files(cls, top_dir):
        """Find all settings.py files in the package

        For that it gets the path to where top_level.txt points to,
        which is not always a folder:
        - folder example: z3c.dependencychecker itself
        - file example: flake8-isort or any other single file distribution

        Return this very same class, which would allow to call the scan()
        method to get an iterator over all this file's imports.
        """
        if top_dir.endswith(".py"):
            return

        for path, folders, filenames in os.walk(top_dir):
            folders[:] = [d for d in folders if d not in FOLDERS_TO_IGNORE]
            for filename in filenames:
                if fnmatch.fnmatch(filename, "*settings.py"):
                    yield cls(
                        top_dir,
                        os.path.join(path, filename),
                    )

    def scan(self):
        for node in ast.walk(self._get_tree()):
            if isinstance(node, ast.Assign):
                if self._is_installed_apps_assignment(node):
                    if isinstance(node.value, (ast.Tuple, ast.List)):
                        for element in node.value.elts:
                            if isinstance(element, ast.Str):
                                yield DottedName(
                                    element.s,
                                    file_path=self.path,
                                    is_test=self.testing,
                                )

                if self._is_test_runner_assignment(node):
                    if isinstance(node.value, ast.Str):
                        yield DottedName(
                            node.value.s,
                            file_path=self.path,
                            is_test=True,
                        )

    @staticmethod
    def _is_installed_apps_assignment(node):
        if (
            len(node.targets) == 1
            and isinstance(node.targets[0], ast.Name)
            and node.targets[0].id == "INSTALLED_APPS"
        ):
            return True

        return False

    @staticmethod
    def _is_test_runner_assignment(node):
        if (
            len(node.targets) == 1
            and isinstance(node.targets[0], ast.Name)
            and node.targets[0].id == "TEST_RUNNER"
        ):
            return True

        return False


MODULES = (
    PythonModule,
    ZCMLFile,
    FTIFile,
    GSMetadata,
    PythonDocstrings,
    DocFiles,
    DjangoSettings,
)



import os

from z3c.dependencychecker.modules import GSMetadata
from z3c.dependencychecker.tests.utils import write_source_file_at

XML_TEMPLATE = """<metadata>
<dependencies>
{0}
</dependencies>
</metadata>
"""

SOMETING_ELSE_IGNORED = """<property class="something"></property>"""
EMPTY_DEPENDENCY = """<dependency></dependency>"""
ONLY_SPACES_DEPENDENCY = """<dependency> </dependency>"""
MISSING_PREFIX = """<dependency>plone.app.caching:default</dependency>"""
MISSING_PROFILE_NAME_SUFFIX = """
<dependency>profile-plone.app.caching</dependency>
"""

ONE_DEPENDENCY = """
<dependency>profile-plone.app.caching:default</dependency>
"""
MORE_DEPENDENCIES = """
<dependency>profile-plone.app.caching:default</dependency>
<dependency>profile-plone.app.dexterity:default</dependency>
"""


def _get_dependencies_on_file(folder, source):
    full_content = XML_TEMPLATE.format(source)
    temporal_file = write_source_file_at(
        (folder.strpath,),
        source_code=full_content,
        filename="configure.zcml",
    )

    gs_metadata = GSMetadata(folder.strpath, temporal_file)
    dotted_names = [x.name for x in gs_metadata.scan()]
    return dotted_names


def test_create_from_files_nothing(minimal_structure):
    path, package_name = minimal_structure
    modules_found = list(GSMetadata.create_from_files(path))
    assert len(modules_found) == 0


def test_create_from_files_deep_nested(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, "src")
    write_source_file_at(
        [src_path, "a", "b", "c"],
        filename="metadata.xml",
    )

    modules_found = list(GSMetadata.create_from_files(src_path))
    assert len(modules_found) == 1


def test_ignore_other_nodes(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, SOMETING_ELSE_IGNORED)
    assert len(dotted_names) == 0


def test_empty_dependency(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, EMPTY_DEPENDENCY)
    assert len(dotted_names) == 0


def test_only_spaces_dependency(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, ONLY_SPACES_DEPENDENCY)
    assert len(dotted_names) == 0


def test_malformed_dependency_no_prefix(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, MISSING_PREFIX)
    assert len(dotted_names) == 0


def test_malformed_dependency_no_suffix(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MISSING_PROFILE_NAME_SUFFIX,
    )
    assert len(dotted_names) == 0


def test_one_dependency(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, ONE_DEPENDENCY)
    assert len(dotted_names) == 1


def test_one_dependency_details(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, ONE_DEPENDENCY)
    assert dotted_names == ["plone.app.caching"]


def test_more_dependencies(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, MORE_DEPENDENCIES)
    assert len(dotted_names) == 2


def test_more_dependencies_details(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, MORE_DEPENDENCIES)
    assert "plone.app.caching" in dotted_names
    assert "plone.app.dexterity" in dotted_names



"""Standard tests configuration module automatically scanned by pytest.

This module is used to add generic fixtures and other boilerplate needed by
our test suite.

Keeping it here ensures that pytest finds it without extra setup and eases
reusability.
"""
import os
import random
import shutil
import string
import tempfile

import pkg_resources
import pytest

from z3c.dependencychecker.dotted_name import DottedName
from z3c.dependencychecker.package import ImportsDatabase


@pytest.fixture
def fake_project():
    temp_folder = tempfile.mkdtemp(prefix="depcheck")
    fake_package_files = pkg_resources.resource_filename(
        "z3c.dependencychecker.tests",
        "sample1",
    )
    package_folder = os.path.join(temp_folder, "sample1")
    shutil.copytree(fake_package_files, package_folder)
    # To prevent the sample .py files to be picked up by ourselves or other
    # tools, I'm postfixing them with ``_in``, now we get to rename them.
    # Same for zcml files.
    for dirpath, dirnames, filenames in os.walk(package_folder):
        for filename in filenames + dirnames:
            if not filename.endswith("_in"):
                continue
            new_filename = filename.replace("_in", "")
            if new_filename == "_it__.py":
                # Oopsie :-) The replace works too well...
                new_filename = "__init__.py"
            source = os.path.join(dirpath, filename)
            target = os.path.join(dirpath, new_filename)
            os.rename(source, target)

    yield package_folder
    shutil.rmtree(temp_folder)


@pytest.fixture
def minimal_structure():
    """Creates a folder structure that contains the minimal files needed
    to make Package class be able to initialize without errors
    """
    folder = tempfile.mkdtemp()
    _add_setup_py(folder)
    package_name = _add_egg_info(folder)

    src_folder = os.path.join(folder, "src")
    os.makedirs(src_folder)

    yield folder, package_name

    shutil.rmtree(folder)


def _add_setup_py(folder):
    with open(os.path.join(folder, "setup.py"), "w") as setup_py_file:
        setup_py_file.write("hi")


def _add_egg_info(folder):
    package_name = "".join(random.choice(string.ascii_lowercase) for _ in range(10))

    egg_info_folder_path = os.path.join(
        folder,
        f"{package_name}.egg-info",
    )
    os.makedirs(egg_info_folder_path)

    _write_pkg_info_file(egg_info_folder_path)
    _write_requires_file(egg_info_folder_path)
    _write_top_level_file(egg_info_folder_path, package_name)

    return package_name


def _write_pkg_info_file(folder):
    with open(os.path.join(folder, "PKG-INFO"), "w") as pkg_info:
        lines = "\n".join(
            ["Metadata-Version: 1.0", "Name: testpackage", "Version: 1.0.dev0"]
        )
        pkg_info.write(lines)


def _write_requires_file(folder):
    with open(os.path.join(folder, "requires.txt"), "w") as requires_file:
        lines = "\n".join(["one", "two"])
        requires_file.write(lines)


def _write_top_level_file(folder_path, package_name):
    file_path = os.path.join(folder_path, "top_level.txt")
    with open(file_path, "w") as top_level_file:
        lines = "\n".join([package_name])
        top_level_file.write(lines)

    sources_top_folder = os.path.join(
        folder_path,
        "..",
        package_name,
    )
    os.makedirs(sources_top_folder)


@pytest.fixture
def minimal_database():
    database = ImportsDatabase()
    database.own_dotted_name = DottedName("fake")
    return database



from unittest import mock

import pytest

from z3c.dependencychecker.package import Package
from z3c.dependencychecker.report import Report
from z3c.dependencychecker.tests.utils import write_source_file_at


def test_print_header_nothing(capsys):
    fake_package = mock.MagicMock()
    report = Report(fake_package)
    report._print_header(None)
    out, err = capsys.readouterr()
    assert out == ""


def test_print_header_text(capsys):
    fake_project = mock.MagicMock()
    report = Report(fake_project)
    report._print_header("hi")
    out, err = capsys.readouterr()

    assert "hi" in out
    assert "==" in out
    assert "===" not in out


def test_print_report(capsys, minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    report = Report(package)
    report.print_report()
    out, err = capsys.readouterr()
    assert "======" not in out  # there are no errors on the dependencies
    assert "" == err


def test_missing_requirements_nothing(capsys):
    fake_project = mock.MagicMock()
    report = Report(fake_project)
    report.missing_requirements()
    out, err = capsys.readouterr()
    assert out == ""


def test_missing_requirements(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at((path, package_name), "__init__.py", "import another.package")

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_requirements()
    out, err = capsys.readouterr()

    assert "Missing requirements\n====================" in out
    assert "another.package" in out


def test_missing_requirements_with_user_mapping(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "import Products.Five.browser.views.BrowserView",
    )
    write_source_file_at(
        (path, f"{package_name}.egg-info"),
        "requires.txt",
        "Zope2",
    )
    write_source_file_at(
        (path,),
        "pyproject.toml",
        "\n".join(
            [
                "[tool.dependencychecker]",
                'Zope2 = ["Four", "Products.Five", "Three" ]',
            ]
        ),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_requirements()
    out, err = capsys.readouterr()

    assert "Missing requirements\n====================" not in out
    assert "Products.Five" not in out
    assert "Zope2" not in out


def test_missing_requirements_with_ignored_packages(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "import Products.Five.browser.views.BrowserView",
    )
    write_source_file_at(
        (path,),
        "pyproject.toml",
        "\n".join(
            [
                "[tool.dependencychecker]",
                'ignore-packages = ["Four", "Products.Five", "Three" ]',
            ]
        ),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_requirements()
    out, err = capsys.readouterr()

    assert "Missing requirements\n====================" not in out
    assert "Products.Five" not in out
    assert "Four" not in out
    assert "Three" not in out


def test_missing_test_requirements(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "",
    )
    write_source_file_at(
        (path, package_name, "tests"),
        "__init__.py",
        "import another.package",
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_test_requirements()
    out, err = capsys.readouterr()

    assert "Missing test requirements\n=========================" in out
    assert "another.package" in out


def test_missing_test_requirements_with_user_mapping(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "",
    )
    write_source_file_at(
        (path, package_name, "tests"),
        "__init__.py",
        "import Products.Five.browser.views.BrowserView",
    )
    write_source_file_at(
        (path, f"{package_name}.egg-info"),
        "requires.txt",
        "Zope2",
    )
    write_source_file_at(
        (path,),
        "pyproject.toml",
        "\n".join(["[tool.dependencychecker]", 'Zope2 = ["Products.Five" ]']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_test_requirements()
    out, err = capsys.readouterr()

    assert "" == out
    assert "Missing requirements\n====================" not in out
    assert "Products.Five" not in out
    assert "Zope2" not in out


def test_missing_test_requirements_with_user_mapping_on_test_extra(
    capsys, minimal_structure
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "",
    )
    write_source_file_at(
        (path, package_name, "tests"),
        "__init__.py",
        "import Products.Five.browser.views.BrowserView",
    )
    write_source_file_at(
        (path, f"{package_name}.egg-info"),
        "requires.txt",
        "\n".join(["[test]", "Zope2"]),
    )
    write_source_file_at(
        (path,),
        "pyproject.toml",
        "\n".join(["[tool.dependencychecker]", 'Zope2 = ["Products.Five" ]']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_test_requirements()
    out, err = capsys.readouterr()

    assert "" == out
    assert "Missing requirements\n====================" not in out
    assert "Products.Five" not in out
    assert "Zope2" not in out


def test_missing_test_requirements_with_ignored_packages(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "",
    )
    write_source_file_at(
        (path, package_name, "tests"),
        "__init__.py",
        "import Products.Five.browser.views.BrowserView",
    )
    write_source_file_at(
        (path,),
        "pyproject.toml",
        "\n".join(
            [
                "[tool.dependencychecker]",
                'ignore-packages = ["Products.Five" ]',
            ]
        ),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_test_requirements()
    out, err = capsys.readouterr()

    assert "" == out
    assert "Missing requirements\n====================" not in out
    assert "Products.Five" not in out


def test_unneeded_requirements(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "import this.package",
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_requirements()
    out, err = capsys.readouterr()

    assert "Unneeded requirements\n=====================" in out
    assert "one" in out
    assert "two" in out


def test_unneeded_requirements_with_ignored_packages(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "import this.package",
    )
    write_source_file_at(
        (path,),
        "pyproject.toml",
        "\n".join(
            [
                "[tool.dependencychecker]",
                'ignore-packages = ["one" ]',
            ]
        ),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_requirements()
    out, err = capsys.readouterr()

    assert "Unneeded requirements\n=====================" in out
    assert "one" not in out
    assert "two" in out


def test_unneeded_requirements_with_user_mapping(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "from BTrees import Tree",
    )
    write_source_file_at(
        (path,),
        "pyproject.toml",
        "\n".join(["[tool.dependencychecker]", '"ZODB3" = ["BTrees" ]']),
    )
    write_source_file_at(
        (path, f"{package_name}.egg-info"),
        "requires.txt",
        "\n".join(["ZODB3", "setuptools", "one"]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_requirements()
    out, err = capsys.readouterr()

    assert "Unneeded requirements\n=====================" in out
    assert "ZODB3" not in out
    assert "one" in out


def test_unneeded_requirements_with_user_mapping2(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "from Plone import Tree",
    )
    write_source_file_at(
        (path,),
        "pyproject.toml",
        "\n".join(["[tool.dependencychecker]", '"ZODB3" = ["BTrees" ]']),
    )
    write_source_file_at(
        (path, f"{package_name}.egg-info"),
        "requires.txt",
        "\n".join(["ZODB3", "setuptools", "one"]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_requirements()
    out, err = capsys.readouterr()

    assert "Unneeded requirements\n=====================" in out
    assert "ZODB3" in out
    assert "one" in out


def test_unneeded_test_requirements(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name + ".egg-info"),
        "requires.txt",
        "\n".join(["[test]", "pytest", "mock"]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_test_requirements()
    out, err = capsys.readouterr()

    assert "Unneeded test requirements\n==========================" in out
    assert "pytest" in out
    assert "mock" in out


def test_unneeded_test_requirements_with_ignore_packages(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name + ".egg-info"),
        "requires.txt",
        "\n".join(["[test]", "pytest", "mock"]),
    )
    write_source_file_at(
        (path,),
        "pyproject.toml",
        "\n".join(["[tool.dependencychecker]", 'ignore-packages = ["mock" ]']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_test_requirements()
    out, err = capsys.readouterr()

    assert "Unneeded test requirements\n==========================" in out
    assert "pytest" in out
    assert "mock" not in out


def test_unneeded_test_requirements_with_user_mappings(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name + ".egg-info"),
        "requires.txt",
        "\n".join(["[test]", "pytest", "ZODB3"]),
    )
    write_source_file_at(
        (path,),
        "pyproject.toml",
        "\n".join(["[tool.dependencychecker]", '"ZODB3" = ["BTrees" ]']),
    )
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "",
    )
    write_source_file_at(
        (path, package_name, "tests"),
        "__init__.py",
        "from BTrees import Tree",
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_test_requirements()
    out, err = capsys.readouterr()

    assert "Unneeded test requirements\n==========================" in out
    assert "pytest" in out
    assert "ZODB3" not in out
    assert "BTrees" not in out


def test_unneeded_test_requirements_no_tests_requirements(
    capsys,
    minimal_structure,
):
    path, package_name = minimal_structure

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_test_requirements()
    out, err = capsys.readouterr()

    assert "" == out


def test_requirements_that_should_be_test_requirements(
    capsys,
    minimal_structure,
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "import one",
    )
    write_source_file_at(
        (path, package_name, "tests"),
        "__init__.py",
        "import two",
    )
    write_source_file_at(
        (path, package_name + ".egg-info"),
        "requires.txt",
        "\n".join(["one", "two", "[test]", "pytest", "mock"]),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.requirements_that_should_be_test_requirements()
    out, err = capsys.readouterr()

    assert (
        "Requirements that should be test requirements\n"
        "=============================================" in out
    )
    assert "two" in out


def test_requirements_that_should_be_test_requirements_with_ignored_packages(
    capsys,
    minimal_structure,
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "import one",
    )
    write_source_file_at(
        (path, package_name, "tests"),
        "__init__.py",
        "\n".join(["import two", "import three"]),
    )
    write_source_file_at(
        (path, package_name + ".egg-info"),
        "requires.txt",
        "\n".join(["one", "two", "three", "[test]", "pytest", "mock"]),
    )
    write_source_file_at(
        (path,),
        "pyproject.toml",
        "\n".join(["[tool.dependencychecker]", 'ignore-packages = ["two" ]']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.requirements_that_should_be_test_requirements()
    out, err = capsys.readouterr()

    assert (
        "Requirements that should be test requirements\n"
        "=============================================" in out
    )
    assert "three" in out
    assert "two" not in out


def test_requirements_that_should_be_test_requirements_with_user_mappings(
    capsys,
    minimal_structure,
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "import one",
    )
    write_source_file_at(
        (path, package_name, "tests"),
        "__init__.py",
        "\n".join(["import two", "import BTrees"]),
    )
    write_source_file_at(
        (path, package_name + ".egg-info"),
        "requires.txt",
        "\n".join(["one", "two", "ZODB3", "[test]", "pytest", "mock"]),
    )
    write_source_file_at(
        (path,),
        "pyproject.toml",
        "\n".join(["[tool.dependencychecker]", '"ZODB3" = ["BTrees" ]']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.requirements_that_should_be_test_requirements()
    out, err = capsys.readouterr()

    assert (
        "Requirements that should be test requirements\n"
        "=============================================" in out
    )
    assert "ZODB3" in out
    assert "two" in out
    assert "one" not in out


def test_requirements_that_should_be_test_requirements_with_user_mapping2(
    capsys,
    minimal_structure,
):
    """Test that user mappings used in both regular and test imports are
    not downgraded to test imports only.
    """
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        "__init__.py",
        "\n".join(["import one", "import two", "import part.of.meta"]),
    )
    write_source_file_at(
        (path, package_name, "tests"),
        "__init__.py",
        "\n".join(["import three", "import four", "import part.of.meta"]),
    )
    write_source_file_at(
        (path, package_name + ".egg-info"),
        "requires.txt",
        "\n".join(["one", "two", "three", "meta-package", "[test]", "four"]),
    )
    write_source_file_at(
        (path,),
        "pyproject.toml",
        "\n".join(["[tool.dependencychecker]", 'meta-package = ["part.of.meta" ]']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.requirements_that_should_be_test_requirements()
    out, err = capsys.readouterr()

    assert (
        "Requirements that should be test requirements\n"
        "=============================================" in out
    )
    assert "three" in out
    assert "one" not in out
    assert "two" not in out
    assert "part.of.meta" not in out
    assert "meta-package" not in out


def test_print_notice(capsys, minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    report = Report(package)
    report.print_notice()
    out, err = capsys.readouterr()
    assert "Note: " in out
    assert "" == err


def test_exit_status_set(minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    package.inspect()
    report = Report(package)
    assert report.exit_status == 0
    report.print_report()
    assert report.exit_status == 1


# These 'lines' are instructions for the parametrized test below. A filename + the desired contents.
PKG_REQUIRES_LINE = ("requires.txt", "zope.interface")
ZOPE_MAP_REQUIRES_LINE = ("requires.txt", "Zope")
MAPPING_LINE = ("pyproject.toml", '[tool.dependencychecker]\nZope = ["zope.interface"]')
CODE_LINE = ("__init__.py", "import zope.interface.Interface")


@pytest.mark.parametrize(
    "files_data,is_missing",
    (
        # no install_requires
        #
        # - no user mapping, with/out code
        ((), False),
        ((CODE_LINE,), True),
        # - user mapping with/out code
        ((MAPPING_LINE,), False),
        ((MAPPING_LINE, CODE_LINE), True),
        #
        # zope.interface in install requires
        #
        # - no user mapping, with/out code
        ((PKG_REQUIRES_LINE,), False),
        ((PKG_REQUIRES_LINE, CODE_LINE), False),
        # - user mapping with/out code
        ((PKG_REQUIRES_LINE, MAPPING_LINE), False),
        ((PKG_REQUIRES_LINE, MAPPING_LINE, CODE_LINE), False),
        #
        # Zope (user mapping) in install requires
        #
        # - no user mapping, with/out code
        ((ZOPE_MAP_REQUIRES_LINE,), False),
        ((ZOPE_MAP_REQUIRES_LINE, CODE_LINE), True),
        # - user mapping with/out code
        ((ZOPE_MAP_REQUIRES_LINE, MAPPING_LINE), False),
        ((ZOPE_MAP_REQUIRES_LINE, MAPPING_LINE, CODE_LINE), False),
    ),
)
def test_missing_requirements_report_zope_interface(
    capsys, minimal_structure, files_data, is_missing
):
    """Test the following cases matrix combinations:

    - on setup.py, one of the following:
      - nothing
      - zope.interface
      - Zope
    - on user mapping, one of the following:
      - nothing
      - Zope = ['zope.interface']
    - on actual code, one of the following:
      - nothing
      - import zope.interface.Interface

    That makes 12 combinations. See the matrix above.
    """
    path, package_name = minimal_structure

    for filename, content in files_data:
        if filename == "requires.txt":
            folder = (path, package_name + ".egg-info")
        elif filename == "pyproject.toml":
            folder = (path,)
        else:
            folder = (path, package_name)

        write_source_file_at(folder, filename, content)

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_requirements()
    out, err = capsys.readouterr()

    if is_missing:
        assert "Missing requirements\n====================" in out
        assert "zope.interface.Interface" in out
    else:
        assert "Missing requirements\n====================" not in out


ZPUB_MAP_LINE = ("pyproject.toml", '[tool.dependencychecker]\nZope = ["ZPublisher"]')
ZPUB_CODE_LINE = ("__init__.py", "import ZPublisher.BaseRequest.RequestContainer")


@pytest.mark.parametrize(
    "files_data,is_missing",
    (
        # no install_requires
        #
        # - no user mapping, with/out code
        ((), False),
        ((ZPUB_CODE_LINE,), True),
        # - user mapping with/out code
        ((ZPUB_MAP_LINE,), False),
        ((ZPUB_MAP_LINE, ZPUB_CODE_LINE), True),
        #
        # Zope (user mapping) in install requires
        #
        # - no user mapping, with/out code
        ((ZOPE_MAP_REQUIRES_LINE,), False),
        ((ZOPE_MAP_REQUIRES_LINE, ZPUB_CODE_LINE), True),
        # - user mapping with/out code
        ((ZOPE_MAP_REQUIRES_LINE, ZPUB_MAP_LINE), False),
        ((ZOPE_MAP_REQUIRES_LINE, ZPUB_MAP_LINE, ZPUB_CODE_LINE), False),
    ),
)
def test_missing_requirements_report_zpublisher(
    capsys, minimal_structure, files_data, is_missing
):
    """Test the following cases matrix combinations:

    - on setup.py, one of the following:
      - nothing
      - Zope
    - on user mapping, one of the following:
      - nothing
      - Zope = ['ZPublisher']
    - on actual code, one of the following:
      - nothing
      - import ZPublisher.BaseRequest.RequestContainer

    That makes 8 combinations. See the matrix above.
    """
    path, package_name = minimal_structure

    for filename, content in files_data:
        if filename == "requires.txt":
            folder = (path, package_name + ".egg-info")
        elif filename == "pyproject.toml":
            folder = (path,)
        else:
            folder = (path, package_name)

        write_source_file_at(folder, filename, content)

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_requirements()
    out, err = capsys.readouterr()

    if is_missing:
        assert "Missing requirements\n====================" in out
        assert "ZPublisher.BaseRequest.RequestContainer" in out
    else:
        assert "Missing requirements\n====================" not in out



import os
import tempfile

import pytest

from z3c.dependencychecker.modules import PythonModule
from z3c.dependencychecker.tests.utils import write_source_file_at

IMPORT = "import foo"
IMPORT_MULTIPLE = "import foo, bar"
IMPORT_AS = "import foo as bar"
IMPORT_AS_MULTIPLE = "import foo as bar, boo as bla"

FROM_IMPORT = "from foo import bar"
FROM_IMPORT_MULTIPLE = "from foo import bar, ber"
FROM_IMPORT_AS = "from foo import bar as ber"
FROM_IMPORT_AS_MULTIPLE = "from foo import bar as ber, boo as bla"
FROM_IMPORT_ASTERISK = "from os import *"
FROM_IMPORT_DOT = "from . import something"
FROM_IMPORT_DOT_DOT = "from .. import something"
FROM_IMPORT_DOT_RELATIVE = "from .local import something"


def _get_imports_of_python_module(folder, source):
    temporal_file = write_source_file_at(
        (folder.strpath,),
        source_code=source,
    )

    python_module = PythonModule(folder.strpath, temporal_file)
    dotted_names = [x.name for x in python_module.scan()]
    return dotted_names


def test_create_from_files_nothing(minimal_structure):
    path, package_name = minimal_structure
    modules_found = list(PythonModule.create_from_files(path))
    assert len(modules_found) == 1


def test_nested_structure_pep420(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, "src")
    write_source_file_at([src_path, "a", "b"], filename="bla.py")
    modules_found = list(PythonModule.create_from_files(src_path))
    assert len(modules_found) == 1


def test_create_from_files_single_file():
    _, tmp_file = tempfile.mkstemp(".py")
    modules_found = list(PythonModule.create_from_files(tmp_file))
    assert len(modules_found) == 1
    assert modules_found[0].path == tmp_file


def test_create_from_files_no_init(minimal_structure):
    """But as we are respecting PEP 420 the python modules are found."""
    path, package_name = minimal_structure
    src_path = os.path.join(path, "src")
    assert len(os.listdir(src_path)) == 0

    write_source_file_at([src_path])
    assert len(os.listdir(src_path)) == 1

    modules_found = list(PythonModule.create_from_files(src_path))
    assert len(modules_found) == 1


def test_create_from_files_ignore_no_python(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, "src")
    write_source_file_at([src_path], filename="__init__.py")
    tempfile.mkstemp(".pt", "bla", src_path)

    modules_found = list(PythonModule.create_from_files(src_path))
    assert len(modules_found) == 1


def test_create_from_files_found_python(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, "src")
    write_source_file_at([src_path], filename="__init__.py")
    write_source_file_at([src_path], filename="bla.py")

    modules_found = list(PythonModule.create_from_files(src_path))
    assert len(modules_found) == 2


def test_create_from_files_deep_nested(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, "src")
    write_source_file_at([src_path], filename="__init__.py")
    write_source_file_at([src_path], filename="bla.py")
    write_source_file_at([src_path, "a"], filename="__init__.py")
    write_source_file_at([src_path, "a"], filename="bla.py")
    write_source_file_at([src_path, "a", "b"], filename="__init__.py")
    write_source_file_at([src_path, "a", "b"], filename="bla.py")

    modules_found = list(PythonModule.create_from_files(src_path))
    assert len(modules_found) == 6


def test_import(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT)
    assert len(dotted_names) == 1


def test_import_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT)
    assert dotted_names == ["foo"]


def test_import_multiple(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT_MULTIPLE)
    assert len(dotted_names) == 2


def test_import_multiple_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT_MULTIPLE)
    assert sorted(dotted_names) == ["bar", "foo"]


def test_import_as(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT_AS)
    assert len(dotted_names) == 1


def test_import_as_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT_AS)
    assert dotted_names == ["foo"]


def test_import_as_multiple(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT_AS_MULTIPLE)
    assert len(dotted_names) == 2


def test_import_as_multiple_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, IMPORT_AS_MULTIPLE)
    assert sorted(dotted_names) == ["boo", "foo"]


def test_from_import(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT)
    assert len(dotted_names) == 1


def test_from_import_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT)
    assert dotted_names == ["foo.bar"]


def test_from_import_multiple(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT_MULTIPLE)
    assert len(dotted_names) == 2


def test_from_import_multiple_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT_MULTIPLE)
    assert sorted(dotted_names) == ["foo.bar", "foo.ber"]


def test_from_import_as(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT_AS)
    assert len(dotted_names) == 1


def test_from_import_as_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT_AS)
    assert dotted_names == ["foo.bar"]


def test_import_asterisk(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT_ASTERISK)
    assert len(dotted_names) == 1


def test_import_asterisk_details(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, FROM_IMPORT_ASTERISK)
    assert dotted_names == ["os"]


@pytest.mark.parametrize(
    "statement",
    [
        FROM_IMPORT_DOT,
        FROM_IMPORT_DOT_DOT,
        FROM_IMPORT_DOT_RELATIVE,
    ],
)
def test_ignore_local_imports(statement, tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, statement)
    assert len(dotted_names) == 0


def test_from_import_as_multiple(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        FROM_IMPORT_AS_MULTIPLE,
    )
    assert len(dotted_names) == 2


def test_from_import_as_multiple_details(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        FROM_IMPORT_AS_MULTIPLE,
    )
    assert sorted(dotted_names) == ["foo.bar", "foo.boo"]


def test_imports_multiple_lines(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        "import foo\nimport bar",
    )
    assert sorted(dotted_names) == ["bar", "foo"]



import codecs
import os.path

from setuptools import find_packages, setup

version = "2.14.1"


def read(filename):
    try:
        with codecs.open(filename, encoding="utf-8") as f:
            return unicode(f.read())
    except NameError:
        with open(filename, encoding="utf-8") as f:
            return f.read()


long_description = f"""
{read('README.rst')}

{read(os.path.join('z3c', 'dependencychecker', 'USAGE.rst'))}

{read('CHANGES.rst')}
"""


setup(
    name="z3c.dependencychecker",
    version=version,
    description="Reports on missing or unneeded python dependencies",
    long_description=long_description,
    long_description_content_type="text/x-rst; charset=UTF-8",
    # Get strings from https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Quality Assurance",
    ],
    keywords="dependencies requirements missing imports",
    author="Reinout van Rees",
    author_email="reinout@vanrees.org",
    url="https://github.com/reinout/z3c.dependencychecker",
    license="BSD",
    packages=find_packages(exclude=["ez_setup"]),
    namespace_packages=["z3c"],
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.8",
    install_requires=[
        "setuptools",
        'stdlib-list; python_version < "3.10"',
        "cached-property",
        "toml",
    ],
    extras_require={
        "test": ["pytest"],
    },
    entry_points={
        "console_scripts": [
            "dependencychecker = z3c.dependencychecker.main:main",
        ]
    },
)



import os

from z3c.dependencychecker.modules import FTIFile
from z3c.dependencychecker.tests.utils import write_source_file_at

XML_TEMPLATE = """<?xml version="1.0"?>
<object>
{0}
</object>
"""

NO_NAME = """
<property class="something"></property>
"""
NO_BEHAVIORS = """
<property name="behaviors"></property>
"""
BEHAVIORS_WITHOUT_VALUE = """
<property name="behaviors">
  <element class="my.behavior"/>
</property>
"""
KLASS = """
<property name="klass">my.class.package</property>
"""
SCHEMA = """
<property name="schema">my.path.to.schema</property>
"""
SOMETHING_ELSE_IGNORE = """
<property name="random_name">my.path.to.schema</property>
"""

SCHEMA_EMPTY = '<property name="schema"></property>'


def _get_fti_imports_on_file(folder, source):
    full_content = XML_TEMPLATE.format(source)
    temporal_file = write_source_file_at(
        (folder.strpath,),
        source_code=full_content,
        filename="configure.zcml",
    )

    fti_file = FTIFile(folder.strpath, temporal_file)
    dotted_names = [x.name for x in fti_file.scan()]
    return dotted_names


def test_create_from_files_nothing(minimal_structure):
    path, package_name = minimal_structure
    modules_found = list(FTIFile.create_from_files(path))
    assert len(modules_found) == 0


def test_create_from_files_deep_nested(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, "src")
    write_source_file_at(
        [src_path, "a", "b", "c", "types"],
        filename="test.xml",
    )

    modules_found = list(FTIFile.create_from_files(src_path))
    assert len(modules_found) == 1


def test_ignore_nodes_without_name(tmpdir):
    dotted_names = _get_fti_imports_on_file(tmpdir, NO_NAME)
    assert len(dotted_names) == 0


def test_no_behavior(tmpdir):
    dotted_names = _get_fti_imports_on_file(tmpdir, NO_BEHAVIORS)
    assert len(dotted_names) == 0


def test_behavior_without_value(tmpdir):
    dotted_names = _get_fti_imports_on_file(tmpdir, BEHAVIORS_WITHOUT_VALUE)
    assert len(dotted_names) == 0


def test_klass(tmpdir):
    dotted_names = _get_fti_imports_on_file(tmpdir, KLASS)
    assert len(dotted_names) == 1


def test_klass_details(tmpdir):
    dotted_names = _get_fti_imports_on_file(tmpdir, KLASS)
    assert "my.class.package" in dotted_names


def test_schema(tmpdir):
    dotted_names = _get_fti_imports_on_file(tmpdir, SCHEMA)
    assert len(dotted_names) == 1


def test_schema_details(tmpdir):
    dotted_names = _get_fti_imports_on_file(tmpdir, SCHEMA)
    assert "my.path.to.schema" in dotted_names


def test_something_to_ignore(tmpdir):
    dotted_names = _get_fti_imports_on_file(tmpdir, SOMETHING_ELSE_IGNORE)
    assert len(dotted_names) == 0


def test_schema_emtpy(tmpdir):
    dotted_names = _get_fti_imports_on_file(tmpdir, SCHEMA_EMPTY)
    assert len(dotted_names) == 0



from z3c.dependencychecker.db import ImportsDatabase
from z3c.dependencychecker.dotted_name import DottedName
from z3c.dependencychecker.tests.utils import (
    get_extras_requirements_names,
    get_requirements_names,
    get_requirements_names_for_extra,
)


def test_no_dependencies():
    database = ImportsDatabase()
    assert isinstance(database._requirements, set)
    assert len(database._requirements) == 0


def test_add_dependencies():
    database = ImportsDatabase()
    database.add_requirements([DottedName("one"), DottedName("two")])

    names = get_requirements_names(database)
    assert len(names) == 2
    assert "one" in names
    assert "two" in names


def test_declared_dependencies_no_duplicates():
    """Check that if there are duplicated requirements,
    they are correctly filtered and only one remains
    """
    database = ImportsDatabase()
    database.add_requirements(
        [DottedName("one"), DottedName("two"), DottedName("three"), DottedName("one")]
    )

    requirements = get_requirements_names(database)
    assert len(requirements) == 3


def test_no_extras():
    database = ImportsDatabase()
    assert len(get_extras_requirements_names(database)) == 0


def test_extras_requirements_key_names():
    database = ImportsDatabase()
    database.add_extra_requirements("test", [DottedName("one")])
    database.add_extra_requirements("plone", [DottedName("two")])

    extras = get_extras_requirements_names(database)
    assert len(extras) == 2
    assert "test" in extras
    assert "plone" in extras


def test_add_extra_dependencies():
    database = ImportsDatabase()
    imports = [
        DottedName("one"),
        DottedName("two"),
    ]
    database.add_extra_requirements("test", imports)

    test_extras = get_requirements_names_for_extra(database, extra="test")
    assert len(test_extras) == 2
    assert "one" in test_extras
    assert "two" in test_extras


def test_warning_extra_declared_twice(caplog):
    """Show a warning if an extra is declared twice on setup.py"""
    database = ImportsDatabase()
    database.add_extra_requirements("test", [DottedName("one")])
    database.add_extra_requirements("test", [DottedName("two")])

    last_log = caplog.records[-1]
    message = 'extra requirement "test" is declared twice'
    assert message == last_log.message


def test_direct_requirements_filtered_on_extras():
    """Check that direct requirements are filtered when added to an extra"""
    database = ImportsDatabase()
    database.add_requirements([DottedName("one"), DottedName("two")])
    database.add_extra_requirements("test", [DottedName("one"), DottedName("three")])

    test_extras = get_requirements_names_for_extra(database, extra="test")
    assert len(test_extras) == 1
    assert "three" in test_extras
    assert "one" not in test_extras


def test_filter_duplicate_extra_requirements():
    """If an extra is declared twice, duplicate requirements are filtered"""
    database = ImportsDatabase()
    database.add_extra_requirements("test", [DottedName("one"), DottedName("three")])
    database.add_extra_requirements("test", [DottedName("two"), DottedName("three")])

    test_extras = get_requirements_names_for_extra(database, extra="test")
    assert len(test_extras) == 3
    assert "one" in test_extras
    assert "two" in test_extras
    assert "three" in test_extras


def test_no_used_imports():
    database = ImportsDatabase()
    assert len(database.imports_used) == 0


def test_add_used_imports():
    database = ImportsDatabase()
    database.own_dotted_name = DottedName("something-else")
    database.add_imports([DottedName("one"), DottedName("two")])
    assert len(database.imports_used) == 2


def test_unique_imports(minimal_database):
    minimal_database.add_imports(
        [
            DottedName("zope.component.ISite"),
            DottedName("zope.component"),
            DottedName("zope.component"),
        ]
    )

    dotted_names = minimal_database._get_unique_imports()
    names = [x.name for x in dotted_names]

    assert len(dotted_names) == 2
    assert "zope.component" in names
    assert "zope.component.ISite" in names


def test_unique_and_sorted_imports(minimal_database):
    minimal_database.add_imports(
        [DottedName("zope.c"), DottedName("zope.a"), DottedName("zope.b")]
    )

    dotted_names = minimal_database._get_unique_imports()

    assert len(dotted_names) == 3
    assert dotted_names[0].name == "zope.a"
    assert dotted_names[1].name == "zope.b"
    assert dotted_names[2].name == "zope.c"


def test_filter_out_known_packages(minimal_database):
    dotted_name = DottedName("bla")
    result = minimal_database._filter_out_known_packages(dotted_name)
    assert result is True


def test_filter_out_known_packages_filtered(minimal_database):
    dotted_name = DottedName("setuptools")
    result = minimal_database._filter_out_known_packages(dotted_name)
    assert result is False


def test_filter_out_known_packages_filtered_nested(minimal_database):
    dotted_name = DottedName("setuptools.my.package")
    result = minimal_database._filter_out_known_packages(dotted_name)
    assert result is False


def test_filter_out_known_packages_filtered_nested_other(minimal_database):
    dotted_name = DottedName("pkg_resources.my.package")
    result = minimal_database._filter_out_known_packages(dotted_name)
    assert result is False


def test_filter_out_testing_imports(minimal_database):
    dotted_name = DottedName("bla", is_test=True)
    result = minimal_database._filter_out_testing_imports(dotted_name)
    assert result is False


def test_filter_out_testing_imports_no_testing_import(minimal_database):
    dotted_name = DottedName("bla", is_test=False)
    result = minimal_database._filter_out_testing_imports(dotted_name)
    assert result is True


def test_filter_out_own_package(minimal_database):
    minimal_database.own_dotted_name = DottedName("zope.component")
    result = minimal_database._filter_out_own_package(DottedName("bla"))
    assert result is True


def test_filter_out_own_package_filtered(minimal_database):
    minimal_database.own_dotted_name = DottedName("zope.component")
    result = minimal_database._filter_out_own_package(
        DottedName("zope.component"),
    )
    assert result is False


def test_filter_out_own_package_if_subpackage(minimal_database):
    minimal_database.own_dotted_name = DottedName("zope.component")
    subpackage = DottedName("zope.component.adapter")
    result = minimal_database._filter_out_own_package(subpackage)
    assert result is False


def test_filter_out_requirements(minimal_database):
    pkg = DottedName("zope.component")
    minimal_database.add_requirements([pkg])
    result = minimal_database._filter_out_requirements(pkg)
    assert result is False


def test_filter_out_std_library(minimal_database):
    pkg = DottedName("os.path.join")
    result = minimal_database._filter_out_python_standard_library(pkg)
    assert result is False


def test_filter_out_requirements_keep_other(minimal_database):
    pkg1 = DottedName("zope.component")
    pkg2 = DottedName("zope.interface")

    minimal_database.add_imports([pkg1, pkg2])
    minimal_database.add_requirements([pkg1])
    dotted_names = minimal_database.get_missing_imports()

    assert len(dotted_names) == 1
    assert pkg2 is dotted_names[0]


def test_filter_out_test_requirements_in_extra(minimal_database):
    dotted_name = DottedName("bli.blu.bla")
    minimal_database.add_extra_requirements(
        "test",
        (DottedName("bla"), DottedName("bli")),
    )
    result = minimal_database._filter_out_test_requirements(dotted_name)
    assert result is False


def test_filter_out_test_requirements_in_extra_variant(minimal_database):
    dotted_name = DottedName("bli.blu.bla")
    minimal_database.add_extra_requirements(
        "tests",
        (DottedName("bla"), DottedName("bli")),
    )
    result = minimal_database._filter_out_test_requirements(dotted_name)
    assert result is False


def test_get_imports_used_filter_subpackage(minimal_database):
    subpkg1 = DottedName("zope.component.adapter")
    subpkg2 = DottedName("zope.component.another.one")

    minimal_database.add_imports([subpkg1, subpkg2])
    minimal_database.add_requirements([DottedName("zope.component")])
    dotted_names = minimal_database.get_missing_imports()

    assert len(dotted_names) == 0


def test_get_imports_used_filter_std_library(minimal_database):
    minimal_database.add_imports(
        [
            DottedName("zope.component"),
            DottedName("os.path.join"),
            DottedName("sys.version_info"),
        ]
    )

    dotted_names = minimal_database.get_missing_imports()

    assert len(dotted_names) == 1
    assert dotted_names[0].name == "zope.component"


def test_get_missing_testing_imports(minimal_database):
    test_import = DottedName("zope.component", is_test=True)
    regular_import = DottedName("zope.interface", is_test=False)
    minimal_database.add_imports([test_import, regular_import])

    dotted_names = minimal_database.get_missing_test_imports()

    assert len(dotted_names) == 1
    assert dotted_names[0] is test_import


def test_get_missing_testing_imports_filter_test_extras(minimal_database):
    test_import1 = DottedName("zope.component", is_test=True)
    test_import2 = DottedName("zope.interface", is_test=True)
    minimal_database.add_imports([test_import1, test_import2])
    minimal_database.add_extra_requirements("test", (test_import1,))

    dotted_names = minimal_database.get_missing_test_imports()

    assert len(dotted_names) == 1
    assert dotted_names[0] is test_import2


def test_get_test_extra(minimal_database):
    dotted_name = DottedName("bla")
    minimal_database.add_extra_requirements("test", (dotted_name,))
    result = minimal_database._get_test_extra()
    assert len(result) == 1
    assert dotted_name in result


def test_get_test_extra_plural(minimal_database):
    dotted_name = DottedName("bla")
    minimal_database.add_extra_requirements("tests", (dotted_name,))
    result = minimal_database._get_test_extra()
    assert len(result) == 1
    assert dotted_name in result


def test_get_test_extra_no_extra(minimal_database):
    dotted_name = DottedName("bla")
    minimal_database.add_extra_requirements("other", (dotted_name,))
    result = minimal_database._get_test_extra()
    assert result == []



import os
import tempfile

from z3c.dependencychecker.modules import DjangoSettings
from z3c.dependencychecker.tests.utils import write_source_file_at

RANDOM_CODE = "from zope.component import adapter"
RANDOM_ASSIGNMENT = "a = 3"
APPS_ASSIGNMENT_TO_STRING = 'INSTALLED_APPS = "random"'
APPS_ASSIGNMENT_TO_TUPLE = 'INSTALLED_APPS = ("random1", "random2", )'
APPS_ASSIGNMENT_TO_LIST = 'INSTALLED_APPS = ["random3", "random4", ]'
APPS_ASSIGNMENT_TO_LIST_MIXED = 'INSTALLED_APPS = ["random5", 3, ]'
TEST_RUNNER_ASSIGNMENT_TO_LIST = 'TEST_RUNNER = ["random6", "random7", ]'
TEST_RUNNER_ASSIGNMENT_TO_STRING = 'TEST_RUNNER = "random8"'


def _get_imports_of_python_module(folder, source):
    temporal_file = write_source_file_at(
        (folder.strpath,),
        source_code=source,
    )

    django_settings = DjangoSettings(folder.strpath, temporal_file)
    dotted_names = [x.name for x in django_settings.scan()]
    return dotted_names


def test_create_from_files_nothing(minimal_structure):
    path, package_name = minimal_structure
    modules_found = list(DjangoSettings.create_from_files(path))
    assert len(modules_found) == 0


def test_create_from_files_single_file_random_name():
    _, tmp_file = tempfile.mkstemp(".py")
    modules_found = list(DjangoSettings.create_from_files(tmp_file))
    assert len(modules_found) == 0


def test_create_from_files_single_file_settings_name(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, "src")
    write_source_file_at([src_path], filename="some_settings.py")
    modules_found = list(DjangoSettings.create_from_files(src_path))
    assert len(modules_found) == 1


def test_create_from_files_deep_nested(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, "src")
    write_source_file_at(
        [src_path, "a", "b", "c"],
        filename="anothersettings.py",
    )
    modules_found = list(DjangoSettings.create_from_files(src_path))
    assert len(modules_found) == 1


def test_random_code(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, RANDOM_CODE)
    assert len(dotted_names) == 0


def test_random_assignment(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, RANDOM_ASSIGNMENT)
    assert len(dotted_names) == 0


def test_apps_assignment_to_string(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_STRING,
    )
    assert len(dotted_names) == 0


def test_apps_assignment_to_tuple(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_TUPLE,
    )
    assert len(dotted_names) == 2


def test_apps_assignment_to_tuple_details(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_TUPLE,
    )
    assert "random1" in dotted_names
    assert "random2" in dotted_names


def test_apps_assignment_to_list(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_LIST,
    )
    assert len(dotted_names) == 2


def test_apps_assignment_to_list_details(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_LIST,
    )
    assert "random3" in dotted_names
    assert "random4" in dotted_names


def test_apps_assignment_to_list_mixed(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_LIST_MIXED,
    )
    assert len(dotted_names) == 1


def test_apps_assignment_to_list_mixed_details(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_LIST_MIXED,
    )
    assert dotted_names[0] == "random5"


def test_runner_assignment_to_list(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        TEST_RUNNER_ASSIGNMENT_TO_LIST,
    )
    assert len(dotted_names) == 0


def test_runner_assignment_to_string(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        TEST_RUNNER_ASSIGNMENT_TO_STRING,
    )
    assert len(dotted_names) == 1


def test_apps_assignment_to_string_mixed_details(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        TEST_RUNNER_ASSIGNMENT_TO_STRING,
    )
    assert dotted_names[0] == "random8"



from z3c.dependencychecker.modules import PythonDocstrings
from z3c.dependencychecker.tests.utils import write_source_file_at

NO_DOC = "class MyClass(object): ..."
INVALID_PYTHON = '''
"""
>>> and some text here that breaks the parser
"""
'''
DOC_IN_MODULE = '''
"""
Random docstring with code to be evaluated.

>>> import zope.component
"""
'''
DOC_IN_FUNCTION = '''
def test():
    """Random docstring with code to be evaluated.

    >>> import zope.component.adapter
    """
'''
DOC_IN_CLASS = '''
class MyClass(object):
    """Random docstring with code to be evaluated.

    >>> from zope.component import utility
    """
'''
DOC_IN_METHOD = '''
class MyClass(object):
    def test(self):
        """Docstring with code to be evaluated.

        >>> import zope.interface
        """
'''
MULTIPLE_IMPORTS_SAME_LINE = '''
class MyClass(object):
    def test(self):
        """Docstring with code to be evaluated.

        >>> from zope.interface import Interface, Attribute
        """
'''
MULTIPLE_IMPORTS_DIFFERENT_LINES = '''
class MyClass(object):
    def test(self):
        """Docstring with code to be evaluated.

        >>> from zope.component import adapter
        >>> from zope.component import utility
        """
'''
MULTIPLE_IMPORTS_DIFFERENT_LINES_WITH_INVALID_CODE_LINES_BETWEEN = '''
class MyClass(object):
    def test(self):
        """Docstring with code to be evaluated.

        >>> from zope.component import adapter
        >>> this should  be skipped and next line processed happily
        >>> from zope.component import utility
        """
'''


def _get_dependencies_on_file(folder, source):
    temporal_file = write_source_file_at(
        (folder.strpath,),
        source_code=source,
    )

    docstring = PythonDocstrings(folder.strpath, temporal_file)
    dotted_names = [x.name for x in docstring.scan()]
    return dotted_names


def test_no_docstring(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, NO_DOC)
    assert len(dotted_names) == 0


def test_invalid_code_ignored(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, INVALID_PYTHON)
    assert len(dotted_names) == 0


def test_docstring_in_module(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_MODULE)
    assert len(dotted_names) == 1


def test_docstring_in_module_details(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_MODULE)
    assert dotted_names == ["zope.component"]


def test_docstring_in_function(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_FUNCTION)
    assert len(dotted_names) == 1


def test_docstring_in_function_details(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_FUNCTION)
    assert dotted_names == ["zope.component.adapter"]


def test_docstring_in_class(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_CLASS)
    assert len(dotted_names) == 1


def test_docstring_in_class_details(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_CLASS)
    assert dotted_names == ["zope.component.utility"]


def test_docstring_in_method(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_METHOD)
    assert len(dotted_names) == 1


def test_docstring_in_method_details(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, DOC_IN_METHOD)
    assert dotted_names == ["zope.interface"]


def test_docstring_multiple_imports_same_line(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_SAME_LINE,
    )
    assert len(dotted_names) == 2


def test_docstring_multiple_imports_same_line_details(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_SAME_LINE,
    )
    assert "zope.interface.Interface" in dotted_names
    assert "zope.interface.Attribute" in dotted_names


def test_docstring_multiple_imports_different_line(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES,
    )
    assert len(dotted_names) == 2


def test_docstring_multiple_imports_same_different_line_details(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES,
    )
    assert "zope.component.adapter" in dotted_names
    assert "zope.component.utility" in dotted_names


def test_docstring_multiple_imports_with_invalid_lines(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES_WITH_INVALID_CODE_LINES_BETWEEN,
    )
    assert len(dotted_names) == 2


def test_docstring_multiple_imports_with_invalid_lines_details(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES_WITH_INVALID_CODE_LINES_BETWEEN,
    )
    assert "zope.component.adapter" in dotted_names
    assert "zope.component.utility" in dotted_names



import os
import tempfile

import pytest

from z3c.dependencychecker.modules import BaseModule
from z3c.dependencychecker.tests.utils import write_source_file_at


def test_module_path():
    obj = BaseModule("/some/path", "/some/path/random/bla")
    assert obj.path == "/some/path/random/bla"


def test_scan_raises():
    obj = BaseModule("/some/path", "/some/path/random/bla")
    with pytest.raises(NotImplementedError):
        obj.scan()


def test_create_from_files_raises():
    with pytest.raises(NotImplementedError):
        BaseModule.create_from_files("/some/path")


def test_has_test_with_prefix_in_path():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, "blatest"))
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing


def test_has_test_in_path():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, "test"))
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing


def test_has_tests_in_path():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, "tests"))
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing


def test_has_test_in_filename():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, "bla"), filename="test.py")
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing


def test_has_tests_in_filename():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, "bla"), filename="tests.py")
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing


def test_has_tests_with_suffix_in_filename():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, "bla"), filename="testsohlala.py")
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing


def test_is_not_a_test_module():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, "bla"), filename="bla.py")
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing is False


def test_parent_folder_is_test_but_module_not():
    """Be sure to ignore the filesystem part and take into account the
    package part

    This means that if I have a package in a folder like:
    /home/gil/testing/my.package
    And a file within it like:
    /home/gil/testing/my.package/src/my/package/configure.zcml
    Despite having 'test' in the path, as it is not part of the package,
    it should be ignored.
    """
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, "test", "bla"), filename="bla.py")
    python_module = BaseModule(os.path.join(folder, "test"), temporal_file)
    assert python_module.testing is False
